const fs = require('fs-extra');
const path = require('path');
const md5 = require('md5');
const { readDir, outputDir, exclude, maxSizeKb, includeExtensions } = require('../src/configure');
const { removeRelativePath, getExtension } = require('../src/utils');

const STATE = {
    missing: 'missing',
    inSync: 'inSync',
    different: 'different',
};

// Execute a callback for each text file in this path, recursively
function eachLocalTextFile(path, fn) {
    fs.readdirSync(path).forEach(entry => {
        if (exclude.includes(entry)) { return; }    // no folders/files from the exclude list
        if (entry[0] === '.') { return; }           // no "hidden" folders or files
        const ext = getExtension(entry);
        const fullEntry = path + '/' + entry; // path.join ?
        const stats = fs.statSync(fullEntry);
        if (stats.isFile()) {
            if (includeExtensions.includes(ext)) {
                let kbSize = stats.size / 1000;
                if (kbSize > maxSizeKb) {
                    console.info(`Too big:   ${kbSize} kb file '${fullEntry}' is larger than the configured limit of ${maxSizeKb} kb. Ignoring.`);
                } else {
                    fn(fullEntry);
                }
            } else {
                // silently skip
                // console.info(`Skipping:  '${fullEntry}' is not one of the known extension types: ${includeExtensions}. Ignoring.`);
            }

        } else if (stats.isDirectory()) {
            eachLocalTextFile(fullEntry, fn);
        }
    });
}

// Get a list of all local text files anywhere in or nested in the current directory.
// You can pass `except` to exclude some files
// When done, calls the callback with an array of file names.
function getAllLocalFiles({ except }, fn) {
    let filenames = [];
    eachLocalTextFile(readDir, Array.prototype.push.bind(filenames));

    filenames = filenames
        .map(removeRelativePath) // node uses paths that include './' like './foo' for lookups, but this confuses s3
        .filter( f => !except.includes(f) ); // exclude files that are already in sync

    fn(filenames);
}

// Given a file name and a string of the contents
// Write it to disk asyncronously
function writeFile(name, stringContents) {
    const fullPath = path.join(outputDir, name);

    if (fs.pathExistsSync(fullPath)) {
        console.log(`Exists:    '${fullPath}'. Overwriting...`);
        // TODO: ? Ask the user's permission to overwrite the local file
    }

    // outputfile is from npm's `fs-extra`. It will create the parent directories if they don't exist
    fs.outputFile(fullPath, stringContents, err => {
        if (err) return console.warn(err);
        console.log(`Written:   '${fullPath}'`);
    });
}

// Given an s3 list object
// return one of 'missing', 'inSync', or 'different'
// Note: 'different' computation relies on S3 using `md5` to implement their `ETag` field.
// If this fails we'll false positive to 'different' too often, but it seems to be the best available solution.
// Becomes more difficult for multipart uploads, but may work ok in this case.
// https://stackoverflow.com/questions/6591047/etag-definition-changed-in-amazon-s3
// Note: LastModified dates don't work, because writing to s3 is an edit, and so is writing from Buffer to local file system
function localStateOf({ Key, ETag }) {
    const filename = path.join(outputDir, Key);

    // Does the file exist locally?
    if (!fs.pathExistsSync(filename)) { return STATE.missing; }

    // Are the contents of the file different?
    const localHash = md5(fs.readFileSync(filename));
    const remoteHash = ETag.replace(/"/g, ''); // remove the quote marks (") that are included by aws for some inexplicable reason.
    if (localHash === remoteHash) {
        return STATE.inSync;
    } else {
        return STATE.different;
    }
}

// Given a filename, read it from the local filesystem syncronously
// Return a { Key, Body } tuple compatible with s3
function readFileReturnKeyBody(filename) {
    const Key = filename;
    const Body = fs.readFileSync(Key);
    return { Key, Body }
}

// Given an s3 List Object (mostly metadata)
// Return true if there is a difference in the local file.
// eg: local file is missing
// eg: local file has different contents.
// function isDifferentLocally(file) {
//     switch (localStateOf(file)) {
//         case STATE.inSync:
//             return false;
//         case STATE.missing:
//         case STATE.different:
//         default:
//             return true;
//     }
// }

module.exports = {
    eachLocalTextFile,
    getAllLocalFiles,
    writeFile,
    localStateOf,
    readFileReturnKeyBody,
};
